﻿/*
 * By David Barrett, Microsoft Ltd. 2012. Use at your own risk.  No warranties are given.
 * 
 * DISCLAIMER:
 * THIS CODE IS SAMPLE CODE. THESE SAMPLES ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND.
 * MICROSOFT FURTHER DISCLAIMS ALL IMPLIED WARRANTIES INCLUDING WITHOUT LIMITATION ANY IMPLIED WARRANTIES OF MERCHANTABILITY OR OF FITNESS FOR
 * A PARTICULAR PURPOSE. THE ENTIRE RISK ARISING OUT OF THE USE OR PERFORMANCE OF THE SAMPLES REMAINS WITH YOU. IN NO EVENT SHALL
 * MICROSOFT OR ITS SUPPLIERS BE LIABLE FOR ANY DAMAGES WHATSOEVER (INCLUDING, WITHOUT LIMITATION, DAMAGES FOR LOSS OF BUSINESS PROFITS,
 * BUSINESS INTERRUPTION, LOSS OF BUSINESS INFORMATION, OR OTHER PECUNIARY LOSS) ARISING OUT OF THE USE OF OR INABILITY TO USE THE
 * SAMPLES, EVEN IF MICROSOFT HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGES. BECAUSE SOME STATES DO NOT ALLOW THE EXCLUSION OR LIMITATION
 * OF LIABILITY FOR CONSEQUENTIAL OR INCIDENTAL DAMAGES, THE ABOVE LIMITATION MAY NOT APPLY TO YOU.
 * */

using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Net;
using System.Xml;
using System.IO;
using System.Security.Cryptography.X509Certificates;
using System.Net.Security;

namespace SOAP_Test_Application
{
    class ClassSOAP
    {
        private string _targetURL="";
        private ICredentials _credentials = null;
        private ClassLogger _logger = null;
        private string _authHeader = "";
        private bool _bypassWebProxy = false;
        private WebHeaderCollection _lastResponseHeaders = null;
        private CookieCollection _responseCookies = null;

        public ClassSOAP(string TargetURL, ICredentials Credential, ClassLogger Logger)
        {
            // Initialise the class
            _targetURL = TargetURL;
            _credentials = Credential;
            _logger = Logger;
        }

        public ClassSOAP(string TargetURL, string Username, string Password, ClassLogger Logger)
        {
            // Initialise the class for basic authentication
            string sAuthInfo = Username + ":" + Password;
            _authHeader = Convert.ToBase64String(System.Text.Encoding.Default.GetBytes(sAuthInfo));
            _targetURL = TargetURL;
            _logger = Logger;
        }

        public string TargetURL
        {
            get { return _targetURL; }
            set { _targetURL = value; }
        }

        public ICredentials Credentials
        {
            get { return _credentials; }
            set { _credentials = value; }
        }

        public bool BypassWebProxy
        {
            get { return _bypassWebProxy; }
            set { _bypassWebProxy = value; }
        }

        public WebHeaderCollection LastResponseHeaders
        {
            get
            {
                return _lastResponseHeaders;
            }
        }

        public CookieCollection ResponseCookies
        {
            get
            {
                return _responseCookies;
            }
        }

        private void LogHeaders(WebHeaderCollection Headers, string Description)
        {
            // Log request headers
            try
            {
                string sHeaders = "";
                foreach (string sHeader in Headers.AllKeys)
                {
                    sHeaders += sHeader + ": " + Headers[sHeader] + Environment.NewLine;
                }
                Log(sHeaders, Description);
            }
            catch { }
        }

        private void LogCookies(CookieCollection Cookies, string Description)
        {
            // Log cookies
            try
            {
                if (Cookies.Count == 0) return;
                string sCookies = "";
                foreach (Cookie cookie in Cookies)
                {
                    sCookies += cookie.ToString() + Environment.NewLine;
                }
                Log(sCookies, Description);
            }
            catch { }
        }

        public string SendRequest(string sRequest, CookieCollection oCookies = null)
        {
            // Send the request and return the response
            
            if (String.IsNullOrEmpty(_targetURL)) return "Error occurred: no target server specified";

            string sResponse = "";
            HttpWebRequest oWebRequest = (HttpWebRequest)WebRequest.Create(_targetURL);
            oWebRequest.Headers.Add("Agent", "SOAPTestApp");
            if (_bypassWebProxy)
                oWebRequest.Proxy = null;

            // Set authentication
            if (!String.IsNullOrEmpty(_authHeader))
            {
                // Force basic authentication
                oWebRequest.Headers["Authorization"] = "Basic " + _authHeader;
            }
            else
                oWebRequest.Credentials = _credentials;

            oWebRequest.ContentType = "text/xml;charset=\"utf-8\"";
            oWebRequest.Accept = "text/xml";
            oWebRequest.Method = "POST";
            XmlDocument oSOAPRequest = new XmlDocument();
            try
            {
                oSOAPRequest.LoadXml(sRequest);
            }
            catch (Exception ex)
            {
                sResponse = "Request was invalid XML: " + ex.Message + "\n\r\n\r" + sRequest;
                Log(sResponse, "Response");
                return sResponse;
            }

            oWebRequest.CookieContainer = new CookieContainer();
            if (!(oCookies == null))
            {
                // Add cookies to the request
                oWebRequest.CookieContainer.Add(oCookies);
                LogCookies(oCookies, "Request Cookies");
            }

            Stream stream = null;
            try
            {
                stream = oWebRequest.GetRequestStream();
            }
            catch (Exception ex)
            {
                // Failed to send request
                sResponse = "Error occurred: " + ex.Message + "\n\r\n\r";
                Log(sResponse, "Response");
                return sResponse;
            }
            if (!string.IsNullOrEmpty(sRequest))
                oSOAPRequest.Save(stream);
            stream.Close();
            LogHeaders(oWebRequest.Headers, "Request Headers");
            Log(oSOAPRequest.OuterXml, "Request");

            IAsyncResult asyncResult = oWebRequest.BeginGetResponse(null, null);
            asyncResult.AsyncWaitHandle.WaitOne();

            WebResponse oWebResponse = null;
            try
            {
                oWebResponse = oWebRequest.EndGetResponse(asyncResult);
                _lastResponseHeaders = oWebResponse.Headers;
                LogHeaders(oWebResponse.Headers, "Response Headers");
                _responseCookies = (oWebResponse as HttpWebResponse).Cookies;
                LogCookies(_responseCookies, "Response Cookies");
            }
            catch (Exception ex)
            {
                if (ex is WebException)
                {
                    WebException wex = ex as WebException;
                    sResponse = "Error occurred: " + wex.Message + "\n\r\n\r";
                    if ( !(wex.Response == null) )
                    {
                        using (StreamReader oReader = new StreamReader(wex.Response.GetResponseStream()))
                        {
                            sResponse += oReader.ReadToEnd();
                        }
                        _lastResponseHeaders = wex.Response.Headers;
                        LogHeaders(wex.Response.Headers, "Response Headers");
                    }
                }
                else
                    sResponse = "Error occurred: " + ex.Message + "\n\r\n\r";
            }
            try
            {
                using (StreamReader oReader = new StreamReader(oWebResponse.GetResponseStream()))
                {
                    sResponse += oReader.ReadToEnd();
                }
            }
            catch { }

            try
            {
                oWebResponse.Close();
            }
            catch { }
            Log(sResponse, "Response");
            return sResponse;
        }

        private void Log(string Details, string Description = "")
        {
            try
            {
                if (_logger == null) return;
                _logger.Log(Details, Description);
            }
            catch (Exception ex)
            {
                System.Windows.Forms.MessageBox.Show(ex.Message, "Failed to log information", MessageBoxButtons.OK, MessageBoxIcon.Warning);
            }
        }
    
    }
}
